MODULE isfparmlt
   !!======================================================================
   !!                       ***  MODULE  isfparmlt  ***
   !! Ice shelf parametrisation module :  update surface ocean boundary condition under ice
   !!                   shelf using an ice shelf melt parametrisation
   !!======================================================================
   !! History :  4.0  ! original code
   !!----------------------------------------------------------------------

   USE isf_oce                  ! ice shelf
   USE isftbl , ONLY: isf_tbl   ! ice shelf depth average
   USE isfutils,ONLY: debug     ! debug subroutine

   USE par_oce                  ! ocean space and time domain
   USE dom_oce                  ! ocean space and time domain
   USE oce    , ONLY: ts        ! ocean dynamics and tracers
   USE phycst , ONLY: rcp, rho0 ! physical constants
   USE eosbn2 , ONLY: eos_fzp   ! equation of state

   USE in_out_manager              ! I/O manager
   USE iom        , ONLY: iom_put  ! I/O library
   USE fldread    , ONLY: fld_read, FLD, FLD_N !
   USE lib_fortran, ONLY: glob_sum !
   USE lib_mpp    , ONLY: ctl_stop !

   IMPLICIT NONE

   PRIVATE

   PUBLIC  isfpar_mlt 


   !! * Substitutions
#  include "domzgr_substitute.h90"   
#  include "do_loop_substitute.h90"
   !!----------------------------------------------------------------------
   !! NEMO/OCE 4.0 , NEMO Consortium (2018)
   !! $Id: sbcisf.F90 10536 2019-01-16 19:21:09Z mathiot $
   !! Software governed by the CeCILL license (see ./LICENSE)
   !!----------------------------------------------------------------------
CONTAINS


   SUBROUTINE isfpar_mlt( kt, Kmm, pqhc, pqoce, pfwf )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE isfpar_mlt  ***
      !!
      !! ** Purpose : Compute Salt and Heat fluxes related to ice_shelf 
      !!              melting and freezing 
      !!
      !! ** Method  :  2 parameterizations are available according
      !!                        1 : Specified melt flux
      !!                        2 : Beckmann & Goose parameterization
      !!
      !!---------------------------------------------------------------------
      INTEGER, INTENT(in) ::   kt   ! ocean time step
      INTEGER, INTENT(in) ::   Kmm  ! ocean time level index
      REAL(wp), DIMENSION(A2D(0)), INTENT(inout) ::   pfwf, pqoce, pqhc  ! fresh water, ice-ocean heat and heat content fluxes
      !!---------------------------------------------------------------------
      !
      ! Choose among the available ice shelf parametrisation
      SELECT CASE ( cn_isfpar_mlt )
      CASE ( 'spe' )    ! specified runoff in depth (Mathiot et al., 2017 in preparation)
         !
         CALL isfpar_mlt_spe(   kt, Kmm, pqhc, pqoce, pfwf )
         !
      CASE ( 'bg03' )    ! Beckmann and Goosse parametrisation 
         !
         CALL isfpar_mlt_bg03(  kt, Kmm, pqhc, pqoce, pfwf )
         !
      CASE ( 'oasis' )
         !
         CALL isfpar_mlt_oasis( kt, Kmm, pqhc, pqoce, pfwf )
         !
      CASE DEFAULT
         CALL ctl_stop('STOP', 'unknown isf melt formulation : cn_isfpar (should not see this)')
      END SELECT
      !
      IF (ln_isfdebug) THEN
         IF(lwp) WRITE(numout,*) ''
         CALL debug( 'isfpar_mlt qhc  :', pqhc (:,:) )
         CALL debug( 'isfpar_mlt qoce :', pqoce(:,:) )
         CALL debug( 'isfpar_mlt fwf  :', pfwf(:,:) )
         IF(lwp) WRITE(numout,*) ''
      END IF
      !
   END SUBROUTINE isfpar_mlt


   SUBROUTINE isfpar_mlt_spe( kt, Kmm, pqhc, pqoce, pfwf )
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE isfpar_mlt_spe  ***
      !!
      !! ** Purpose : prescribed ice shelf melting in case ice shelf cavities are closed.
      !!              data read into a forcing files.
      !!
      !!--------------------------------------------------------------------
      INTEGER,  INTENT(in) ::   kt
      INTEGER,  INTENT(in) ::   Kmm    !  ocean time level index
      REAL(wp), DIMENSION(A2D(0)), INTENT(inout) ::   pqhc, pfwf, pqoce  ! fresh water and ice-ocean heat fluxes
      !!--------------------------------------------------------------------
      INTEGER ::   ji, jj, jk     ! dummy loop indices
      REAL(wp), DIMENSION(A2D(0),jpk) ::   ztfrz3d, ztmp
      REAL(wp), DIMENSION(A2D(0))     ::   ztfrz
      !!--------------------------------------------------------------------
      !
      ! 0. ------------Read specified fwf from isf to oce
      CALL fld_read ( kt, 1, sf_isfpar_fwf   )
      !
      ! compute ptfrz
      ! 1. ------------Mean freezing point
      DO_3D( 0, 0, 0, 0, 1, jpk )
         ztmp(ji,jj,jk) = gdept(ji,jj,jk,Kmm)
      END_3D
      CALL eos_fzp( ts(A2D(0),:,jp_sal,Kmm), ztfrz3d(:,:,:), ztmp )
      !
      CALL isf_tbl( Kmm, ztfrz3d, 'T', misfkt_par, rhisf_tbl_par, & ! <<== in
         &                                                 ztfrz, & ! ==>> out
         &                             misfkb_par, rfrac_tbl_par  ) ! <<== in (optional)
      !
      DO_2D( 0, 0, 0, 0 )
         pfwf(ji,jj) =   sf_isfpar_fwf(1)%fnow(ji,jj,1)        ! fresh water flux from the isf (fwfisf <0 mean melting)       ( > 0 from isf to oce)
         pqoce(ji,jj) = - pfwf(ji,jj) * rLfusisf               ! ocean/ice shelf flux assume to be equal to latent heat flux  ( > 0 from isf to oce)
         pqhc (ji,jj) =   pfwf(ji,jj) * ztfrz(ji,jj) * rcp     ! heat content flux                                            ( > 0 from isf to oce)
      END_2D
      !
      CALL iom_put('isftfrz_par', ztfrz(:,:) * mskisf_par(:,:) )
      !
   END SUBROUTINE isfpar_mlt_spe

   SUBROUTINE isfpar_mlt_bg03(kt, Kmm, pqhc, pqoce, pfwf)
      !!---------------------------------------------------------------------
      !!                  ***  ROUTINE isfpar_mlt_bg03  ***
      !!
      !! ** Purpose : compute an estimate of ice shelf melting and 
      !!              latent, ocean-ice and heat content heat fluxes
      !!              in case cavities are closed based on the far fields T and S properties. 
      !!
      !! ** Method  : The ice shelf melt is computed as proportional to the differences between the 
      !!              mean temperature and mean freezing point in front of the ice shelf averaged 
      !!              over the ice shelf min ice shelf draft and max ice shelf draft and the freezing point
      !!
      !! ** Reference : Beckmann and Goosse (2003), "A parameterization of ice shelf-ocean
      !!                interaction for climate models", Ocean Modelling 5(2003) 157-170.
      !!----------------------------------------------------------------------
      INTEGER,  INTENT(in) ::   kt
      INTEGER,  INTENT(in) ::   Kmm    !  ocean time level index
      REAL(wp), DIMENSION(A2D(0)), INTENT(inout) ::   pqhc, pfwf, pqoce  ! fresh water and ice-ocean heat fluxes
      !!--------------------------------------------------------------------
      INTEGER ::   ji, jj, jk     ! dummy loop indices
      REAL(wp), DIMENSION(A2D(0),jpk) ::   ztfrz3d, ztmp
      REAL(wp), DIMENSION(A2D(0))     ::   ztfrz          ! freezing point
      REAL(wp), DIMENSION(A2D(0))     ::   ztavg          ! temperature avg
      !!----------------------------------------------------------------------
      !
      ! 0. ------------Mean freezing point
      DO_3D( 0, 0, 0, 0, 1, jpk )
         ztmp(ji,jj,jk) = gdept(ji,jj,jk,Kmm)
      END_3D
      CALL eos_fzp( ts(A2D(0),:,jp_sal,Kmm), ztfrz3d(:,:,:), ztmp )

      CALL isf_tbl( Kmm, ztfrz3d, 'T', misfkt_par, rhisf_tbl_par, & ! <<== in
         &                                                 ztfrz, & ! ==>> out
         &                             misfkb_par, rfrac_tbl_par  ) ! <<== in (optional)
      !
      ! 1. ------------Mean temperature
      CALL isf_tbl( Kmm, ts(:,:,:,jp_tem,Kmm), 'T', misfkt_par, rhisf_tbl_par, & ! <<== in
         &                                                              ztavg, & ! ==>> out
         &                                          misfkb_par, rfrac_tbl_par  ) ! <<== in (optional)
      !
      ! 2. ------------Net heat flux and fresh water flux due to the ice shelf
      DO_2D( 0, 0, 0, 0 )
         pfwf(ji,jj) =   rho0 * rcp * rn_isfpar_bg03_gt0 * risfLeff(ji,jj) * e1t(ji,jj) * ( ztavg(ji,jj) - ztfrz(ji,jj) ) &
            &                                                               * r1_e1e2t(ji,jj) / rLfusisf  ! ( > 0 from isf to oce)
         pqoce(ji,jj) = - pfwf(ji,jj) * rLfusisf               ! ocean/ice shelf flux assume to be equal to latent heat flux  ( > 0 from isf to oce)
         pqhc (ji,jj) =   pfwf(ji,jj) * ztfrz(ji,jj) * rcp     ! heat content flux                                            ( > 0 from isf to oce)
      END_2D
      !
      ! 3. ------------BG03 output
      CALL iom_put('ttbl_par',         ztavg(:,:)                * mskisf_par(:,:) )      ! ttbl
      CALL iom_put('isfthermald_par',( ztavg(:,:) - ztfrz(:,:) ) * mskisf_par(:,:) )      ! thermal driving
      CALL iom_put('isftfrz_par',                   ztfrz(:,:)   * mskisf_par(:,:) )      ! freezing point

      !
   END SUBROUTINE isfpar_mlt_bg03

   SUBROUTINE isfpar_mlt_oasis( kt, Kmm, pqhc , pqoce, pfwf )
      !!----------------------------------------------------------------------
      !!                  ***  ROUTINE isfpar_mlt_oasis  ***
      !!
      !! ** Purpose    : scale the fwf read from input file by the total amount received by the sbccpl interface
      !!
      !! ** Purpose    : - read ice shelf melt from forcing file and scale it by the input file total amount => pattern
      !!                 - compute total amount of fwf given by sbccpl (fwfisf_oasis)
      !!                 - scale fwf and compute heat fluxes
      !!
      !!---------------------------------------------------------------------
      INTEGER                    , INTENT(in ) ::   kt                  ! current time step
      INTEGER                    , INTENT(in ) ::   Kmm                 !  ocean time level index
      REAL(wp), DIMENSION(A2D(0)), INTENT(out) ::   pqhc, pqoce, pfwf  ! heat content, latent heat and fwf fluxes
      !!--------------------------------------------------------------------
      INTEGER  ::   ji, jj, jk           ! dummy loop indices
      REAL(wp) ::   zfwf_fld, zfwf_oasis ! total fwf in the forcing fields (pattern) and from the cpl interface (amount)
      REAL(wp), DIMENSION(A2D(0))     ::   ztfrz                         ! tbl freezing temperature
      REAL(wp), DIMENSION(A2D(0))     ::   zfwf                          ! 2d fwf map after scaling
      REAL(wp), DIMENSION(A2D(0),jpk) ::   ztfrz3d, ztmp
      !!--------------------------------------------------------------------
      !
      ! 0. ------------Read specified runoff
      CALL fld_read ( kt, 1, sf_isfpar_fwf   )
      !
      ! 1. ------------Mean freezing point (needed for heat content flux)
      DO_3D( 0, 0, 0, 0, 1, jpk )
         ztmp(ji,jj,jk) = gdept(ji,jj,jk,Kmm)
      END_3D
      CALL eos_fzp( ts(A2D(0),:,jp_sal,Kmm), ztfrz3d(:,:,:), ztmp )
      !
      CALL isf_tbl( Kmm, ztfrz3d, 'T', misfkt_par, rhisf_tbl_par, & ! <<== in
         &                                                 ztfrz, & ! ==>> out
         &                             misfkb_par, rfrac_tbl_par  ) ! <<== in (optional)
      !
      ! 2. ------------Scale isf melt pattern with total amount from oasis
      ! ice shelf 2d map of fwf from isf to oce
      zfwf(:,:) = sf_isfpar_fwf(1)%fnow(:,:,1)
      !
      ! compute glob sum from input file
      ! (PM) should we consider delay sum as in fwb ? (it will offset by 1 time step if I understood well)
      zfwf_fld = glob_sum('isfcav_mlt', e1e2t(A2D(0)) * zfwf(:,:))
      !
      ! compute glob sum from atm->oce ice shelf fwf
      ! (PM) should we consider delay sum as in fwb ?
      zfwf_oasis = glob_sum('isfcav_mlt', e1e2t(A2D(0)) * fwfisf_oasis(:,:))
      !
      ! scale fwf
      zfwf(:,:) = zfwf(:,:) * zfwf_oasis / zfwf_fld
      ! 
      ! 3. -----------Define fwf and qoce
      ! ocean heat flux is assume to be equal to the latent heat
      DO_2D( 0, 0, 0, 0 )
         pfwf(ji,jj) =   zfwf(ji,jj)                           ! fwf                ( > 0 from isf to oce)
         pqoce(ji,jj) = - pfwf(ji,jj) * rLfusisf               ! ocean heat flux    ( > 0 from isf to oce) (assumed to be the latent heat flux)
         pqhc (ji,jj) =   pfwf(ji,jj) * ztfrz(ji,jj) * rcp     ! heat content flux  ( > 0 from isf to oce)
      END_2D
      !
      CALL iom_put('isftfrz_par', ztfrz(:,:) * mskisf_par(:,:) )
      !
   END SUBROUTINE isfpar_mlt_oasis

END MODULE isfparmlt
